__version__ = '0.1.4'


from .Freezable import Freezable
from .Accession import Accession
from .Cohort    import Cohort
from .CloudData import CloudData


# Initialize the Minus80 Freezer
def _init():
    import os
    from .Config import cf
 
    basedir = os.path.realpath(
        os.path.expanduser(cf.options.basedir)
    )
    
    # Create the basedir if not exists
    try:
        os.makedirs(basedir, exist_ok=True)
        os.makedirs(os.path.join(basedir, "databases"), exist_ok=True)
        os.makedirs(os.path.join(basedir, "tmp"), exist_ok=True)
        os.makedirs(os.path.join(basedir, "Raw"), exist_ok=True)
    except Exception:
        raise

_init()



#!/usr/env/python3

import os
import yaml
import pprint
import getpass

global cf

default_config = '''--- # YAML Camoco Configuration File
options:
    basedir: ~/.minus80/

cloud:
    endpoint: https://s3.msi.umn.edu
    access_key: None
    secret_key: None

'''.format(**{'user':getpass.getuser()})

class Level(dict):
    '''
        Ha! Take that config parser! I am accessing
        everything like an object.
    '''
    def __init__(self,*args,**kwargs):
        # Create the dict
        super().__init__(*args,**kwargs)
        # Convert to Levels

        for k,v in self.items():
            if isinstance(v,dict):
                self[k] = Level(v)
            else:
                self[k] = v

    def __getattr__(self,item):
        if 'dir' in item and '~' in self[item]:
            return os.path.expanduser(self[item])
        return self[item]
    def __setattr__(self,item,val):
        self[item] = val



class Config(object):

    def __init__(self,filename):
        filename = os.path.expanduser(filename)
        self.data = Level(yaml.safe_load(open(filename,'r')))

    def __getattr__(self,item):
        return self.data[item]

    def __getitem__(self,item):
        return self.data[item]

    def __repr__(self):
        return pprint.pformat(self.data)

#  -------------------------------------------------------------------------
#        Program Logic


cf_file = os.path.expanduser('~/.minus80.conf')

# Check to see if there is a config file available
if not os.path.isfile(cf_file):
    with open(cf_file, 'w') as CF:
        print(default_config, file=CF)

cf = Config(cf_file)



from .Tools import get_files
from .Config import cf

from collections import defaultdict
import os
import lzma
import sys


__all__ = ['CloudData']


def CloudData(engine='s3'):
    if engine == 's3':
        return S3CloudData()
    else:
        raise ValueError(f'Cannot use {engine} as a cloud engine.')


class BaseCloudData(object):
    def __init__(self):
        pass

    def put(self, name, dtype, raw=False, compress=False):
        pass

    def get(self, name, dtype, raw=False):
        pass

    def list(self, name=None, dtype=None, raw=None):
        pass


class S3CloudData(BaseCloudData):

    '''
    CloudData objects allow minus80 to interact with the cloud to store both
    prepared as well as raw datasets.
    '''

    bucket = 'minus80'

    def __init__(self):
        '''
        Create a CloudData object. Once proper S3 credentials are stored in the
        config file (~/.minus80.conf) initialization takes no arguments.
        '''
        try:
            import boto3
            from botocore.client import Config
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                'boto3 must be installed to use this feature'
            )

        if cf.cloud.access_key == 'None' or cf.cloud.secret_key == 'None':
            raise ValueError('Fill in your S3 Credentials in ~/.minus80.conf')

        self.s3 = boto3.client(
            service_name='s3',
            endpoint_url=cf.cloud.endpoint,
            aws_access_key_id=cf.cloud.access_key,
            aws_secret_access_key=cf.cloud.secret_key,
            config=Config(s3={'addressing_style': 'path'})
        )

        # make sure the minus80 bucket exists
        if self.bucket not in [x['Name'] for x in self.s3.list_buckets()['Buckets']]:
            self.s3.create_bucket(Bucket='minus80')


    def put(self, name, dtype, raw=False, compress=False):
        '''
        Store a minus80 dataset in the cloud using its name and dtype (e.g. Cohort).
        the dtype is the name of the Freezable class or object. See :ref:`freezable`.
        Assume we are storing ``x = Cohort('experiment1')``

        name : str
            The name of the dataset (i.e. 'experiment1')
        dtype : str
            The type of freezable object (i.e. 'Cohort')
        raw : bool, default=False
            If True, raw files can be stored in the cloud. In this case, name changes
            to the file name and dtype changes to a string representing the future dtype
            or anything that describes the type of data that is being stored.
        lzma : bool, default=False
            If true, lzma compression will be performed (this is slower)
        '''
        key = os.path.basename(name)
        if raw == True:
            # The name is a FILENAME
            filename = name
            if compress:
                with open(filename, 'rb') as OUT:
                    self.s3.upload_fileobj(lzma.compress(OUT.read()), self.bucket, f'Raw/{dtype}/{key}.xz')
            else:
                self.s3.upload_file(filename, self.bucket, f'Raw/{dtype}/{key}')
        else:
            files = get_files(name, dtype, fullpath=True)
            if len(files) == 0:
                raise ValueError('There were no datasets with that name')
            for filename in files:
                self.s3.upload_file(filename, self.bucket, f'databases/{dtype}/{key}')

    def get(self, name, dtype, raw=False):
        '''
        Retrive a minus80 dataset in the cloud using its name and dtype (e.g. Cohort).
        the dtype is the name of the Freezable class or object. See :ref:`freezable`.
        Assume we are storing ``x = Cohort('experiment1')``

        name : str
            The name of the dataset (i.e. 'experiment1')
        dtype : str
            The type of freezable object (i.e. 'Cohort')
        raw : bool, default=False
            If True, raw files can be stored in the cloud. In this case, name changes
            to the file name and dtype changes to a string representing the future dtype
            or anything that describes the type of data that is being stored.
        lzma : bool, default=False
            If true, lzma compression will be performed (this is slower)
        '''

        # Define a helper
        def get_percent_done(current, total):
            write, flush = sys.stdout.write, sys.stdout.flush
            percent = int((current/total)*100)
            write('\x08'*17)
            write(f'Percent done: {percent}%')
            flush()

        key = os.path.basename(name)
        if raw == True:
            # filename = name
            bdir = os.path.expanduser(cf.options.basedir)
            # get the number of bytes in the object
            num_bytes = self.s3.list_objects(Bucket=self.bucket, Prefix=f'Raw/{dtype}/{key}')['Contents'][0]['Size']
            # download
            os.makedirs(f'{bdir}/Raw/{dtype}')
            with open(f'{bdir}/Raw/{dtype}/{key}', 'wb') as OUT:
                self.s3.download_fileobj(
                    self.bucket,
                    f'Raw/{dtype}/{key}',
                    OUT,
                    Callback = lambda x: get_percent_done(x, num_bytes)
                )
        else:
            raise NotImplementedError('This functionality is currently only available for raw data')

    def list(self, name=None, dtype=None, raw=False):
        items = defaultdict(list)
        try:
            for item in self.s3.list_objects(Bucket=self.bucket)['Contents']:
                key = item['Key']
                if key.startswith('Raw') and raw == False:
                    pass
                elif not key.startswith('Raw') and raw == True:
                    pass
                else:
                    _, key_dtype, key_name = key.split('/')
                    if dtype != None and key_dtype != dtype:
                        pass
                    elif name != None and not key_name.startswith(name):
                        pass
                    else:
                        items[key_dtype].append(key_name)
            if len(items) == 0:
                print('Nothing here yet!')
            else:
                if raw:
                    print('######   Raw Data:   ######')
                for key, vals in items.items():
                    print(f'-----{key}------')
                    print('\n'.join(vals))
        except KeyError:
            if len(items) == 0:
                print('Nothing here yet!')



#!/usr/bin/env python3
import tempfile
import re
# Suppress the warning until the next wersion
import bcolz as bcz

import apsw as lite
import os as os
import numpy as np
import pandas as pd

from .Config import cf
from contextlib import contextmanager


class Freezable(object):

    '''
    Freezable is an abstract class. Things that inherit from Freezable can
    be loaded and unloaded from the Minus80.

    A freezable object is a persistant object that lives in a known directory
    aimed to make expensive to build objects and databases loadable from
    new runtimes.

    The three main things that a Freezable object supplies are:
    * access to a sqlite database (relational records)
    * access to a bcolz databsase (columnar data)
    * access to a key/val store
    * access to named temp files

    '''

    def __init__(self, name, dtype=None, basedir=None):
        '''
        Initialize the Freezable Object.

        Parameters
        ----------
        name : str
            The name of the frozen object.
        dtype : str, default=None
            The type of the frozen object (e.g. Cohort)
            If None, the type will be inferred from the
            object class.
        basedir : str, default=None
            The basedir to store the frozen object. If
            None, the basedire in the config file will be
            used.

        '''
        # Set up our base directory
        if basedir is None:
            self._m80_basedir = cf.options.basedir
        self._m80_name = name
        #
        if dtype is None:
            # Just use the class type as the type
            self._m80_type = self.guess_type(self)
        else:
            self._m80_type = dtype
        # A dataset already exists, return it
        self._db = self._open_db(self._m80_name)

        try:
            cur = self._db.cursor()
            cur.execute('''
                CREATE TABLE IF NOT EXISTS globals (
                    key TEXT,
                    val TEXT,
                    type TEXT
                );
                CREATE UNIQUE INDEX IF NOT EXISTS uniqkey ON globals(key)
                ''')
        except TypeError:
            raise TypeError('{}.{} does not exist'.format(dtype, name))

    @contextmanager
    def bulk_transaction(self):
        '''
            This is a context manager that handles bulk transaction.
            i.e. this context will handle the BEGIN, END and appropriate
            ROLLBACKS.

            Usage:
            >>> with x.bulk_transaction() as cur:
                     cur.execute('INSERT INTO table XXX VALUES YYY')
        '''
        cur = self._db.cursor()
        cur.execute('PRAGMA synchronous = off')
        cur.execute('PRAGMA journal_mode = memory')
        cur.execute('SAVEPOINT bulk_transaction')
        try:
            yield cur
        except Exception as e:
            cur.execute('ROLLBACK TO SAVEPOINT bulk_transaction')
            raise e
        finally:
            cur.execute('RELEASE SAVEPOINT bulk_transaction')

    def _dbfilename(self, dbname=None, dtype=None):
        '''
        Get the path to a database file.

        Parameters
        ----------
        dbname : str, default=None
            The frozen object name
        dtype : str, default=None
            The datatype of the frozen object

        Returns
        -------



        '''
        if dbname is None:
            dbname = self._m80_name
        if dtype is None:
            dtype = self._m80_type
        return os.path.expanduser(
            os.path.join(
                self._m80_basedir,
                'databases',
                '{}.{}.db'.format(dbname, dtype)
            )
        )

    def _open_db(self, dbname=None, dtype=None):
        '''
            This is the access point to the sqlite database
        '''
        # return a connection if exists
        return lite.Connection(
            self._dbfilename(dbname, dtype)
        )

    def _bcolz_array(self, name, array=None, m80name=None,
                     m80type=None):
        '''
            Routines to set/get arrays from the bcolz store
        '''
        # Fill in the defaults if they were not provided
        if m80type is None:
            m80type = self._m80_type
        if m80name is None:
            m80name = self._m80_name
        # function is a getter if df is provided
        path = os.path.expanduser(
            os.path.join(
                self._m80_basedir,
                'databases',
                "{}.{}.bcz".format(m80name, m80type)
            )
        )
        os.makedirs(path, exist_ok=True)
        if array is None:
            # GETTER
            arr = bcz.open(os.path.join(path, name))
            return arr
        else:
            # SETTER
            bcz.carray(array, mode='w', rootdir=os.path.join(path, name))

    def _bcolz(self, tblname, df=None, m80name=None, m80type=None,
               blaze=False):
        '''
            This is the access point to the bcolz database
        '''
        try:
            import blaze as blz
        except FutureWarning:
            pass
        import warnings
        # from flask.exthook import ExtDeprecationWarning
        # warnings.simplefilter('ignore', ExtDeprecationWarning)
        warnings.simplefilter('ignore', FutureWarning)

        # Fill in the defaults if they were not provided
        if m80type is None:
            m80type = self._m80_type
        if m80name is None:
            m80name = self._m80_name
        path = os.path.expanduser(
            os.path.join(
                self._m80_basedir,
                'databases',
                "{}.{}.bcz".format(m80name, m80type)
            )
        )
        os.makedirs(path, exist_ok=True)

        # function is a getter if df is provided
        if df is None:
            # return the dataframe if it exists
            try:
                df = bcz.open(os.path.join(path, tblname))
            except IOError:
                raise IOError(
                    f'could not open database for {m80type}:{m80name} '
                )
            else:
                if len(df) == 0:
                    df = pd.DataFrame()
                    if blaze:
                        df = blz.data(df)
                else:
                    if blaze:
                        df = blz.data(df)
                    else:
                        df = df.todataframe()
                if not blaze and 'idx' in df.columns.values:
                    df.set_index('idx', drop=True, inplace=True)
                    df.index.name = None
                return df
        # If df is set, then store the table
        else:
            if not(df.index.dtype_str == 'int64') and not (df.empty):
                df = df.copy()
                df['idx'] = df.index
            if isinstance(df, pd.DataFrame):
                path = os.path.join(path, tblname)
                if df.empty:
                    bcz.fromiter(
                        (), dtype=np.int32, mode='w',
                        count=0, rootdir=path
                    )
                else:
                    bcz.ctable.fromdataframe(df, mode='w', rootdir=path)
            if 'idx' in df.columns.values:
                del df
            return

    @staticmethod
    def _tmpfile(*args, **kwargs):
        # returns a handle to a tmp file
        return tempfile.NamedTemporaryFile(
            'w',
            dir=os.path.expanduser(
                os.path.join(
                    cf.options.basedir,
                    "tmp"
                )
            ),
            **kwargs
        )

    def _dict(self, key, val=None):
        '''
            Stores global variables for the freezable object. The
            method will automatically infer in the val type is in
            [int, float, str]. If the value is not one of these, an
            excpetion will be raised.

            Parameters
            ----------
            key : str
                the dictionary key
            val : int, float, or str
                the value corresponding to the key

        '''
        try:
            if val is not None:
                val_type = self.guess_type(val)
                if val_type not in ('int', 'float', 'str'):
                    raise TypeError(
                        f'val must be in [int, float, str], not {val_type}'
                    )
                self._db.cursor().execute(
                    '''
                    INSERT OR REPLACE INTO globals
                    (key, val, type)VALUES (?, ?, ?)''', (key, val, val_type)
                )
            else:
                (valtype, value) = self._db.cursor().execute(
                    '''SELECT type, val FROM globals WHERE key = ?''', (key, )
                ).fetchone()
                if valtype == 'int':
                    return int(value)
                elif valtype == 'float':
                    return float(value)
                elif valtype == 'str':
                    return str(value)
        except TypeError:
            raise ValueError('{} not in database'.format(key))

    def _cassandra(self, name=None, dtype=None):
        '''
            Provides an interface to a cassandra NOSQL database. This is
            experimental.
        '''
        try:
            from cassandra.cluster import Cluster
        except ImportError as e:
            raise ImportError(
                'Please install `cassandra-driver` to use this feature', e
            )
        if dtype is None:
            dtype = self._m80_type
        if name is None:
            name = self._m80_name
        cluster = Cluster()
        # Connect to the keyspace dictated by the object
        session = cluster.connect()  # f'{name}.{dtype}')
        return session

    @staticmethod
    def guess_type(object):
        '''
            Guess the type of object from the class attribute
        '''
        # retrieve a list of classes
        classes = re.match(
            "<class '(.+)'>",
            str(object.__class__)
        ).groups()[0].split('.')
        # Return the most specific one
        return classes[-1]



import os
import getpass
import socket
import urllib
import asyncio
import asyncssh
import os

from contextlib import contextmanager

from .Config import cf

class Accession(object):
    '''
    From google: Definition (noun): a new item added to an existing collection
    of books, paintings, or artifacts.

    An Accession is an item that exists in an experimental collection.

    Most of the time an accession is interoperable with a *sample*. However,
    the term sample can become confusing when an experiment has multiple
    samplings from the same sample, e.g. timecourse or different tissues.
    '''

    def __init__(self, name, files=None, **kwargs):
        '''
        Create a new accession.

        Parameters
        ----------
        name : str
            The name of the accession
        files : iterable of str
            Files associated with the accession
        **kwargs : keyword arguments
            Any number of key=value arguments that
            contain metadata.

        Returns
        -------
        An accession object
        '''
        self.name = name
        if files is not None:
            self.files = set(files)
        else:
            self.files = set()
        self.metadata = kwargs

    def __getitem__(self, key):
        '''
        Retrieve metadata about an accession.

        Parameters
        ----------
        key : str

        Returns
        -------
        Value from the accession metadata corresponding
        to the key.
        '''
        return self.metadata[key]

    def __setitem__(self, key, val):
        '''
        Set metadata about an accession

        Parameters
        ----------
        key : str
            The metadata name
        val : str
            The value of the metadata
        '''
        self.metadata[key] = val

    def add_file(self, path, skip_check=False, scheme='ssh',
                 username=None, hostname=None):
        '''
        Add a file that is associated with the accession.
        This method will attempt to determine where the file
        is actually stored based on its path. Currently it 
        supports three different protocols: local, ssh and
        s3. A local file will looks something like:
        `/tmp/file1.fastq`.  

        Parameters
        ----------
        path/URL: string
            The path/URL the the file. The string is parsed
            for default information (e.g. 
        skip_check : bool
            If true, the method will not test if the file
            exists
        scheme: string (default: ssh)
            Specifies the scheme/protocol for accessing the file.
            Defaults to ssh, also supports s3
        username : string (default: None)
            Defines a username that is authorized to access
            `hostname` using `protocol`. Defaults to None 
            in which case it will be determined by calling
            `getpass.getuser()`.
        hostname : sting (default: None)
            Defines the ostname that the file is accessible
            through. Defaults to None, where the hostname
            will be determined 
        port: int (default: 22)
            Port to access the file through. Defaults to 22,
            which is for ssh.

        NOTE: any keyword arguments passed in will override
              the values parsed out of the path. 

        Returns
        -------
        None
        '''
        url = urllib.parse.urlparse(path)
        # Override parsed url values with keywords
        if scheme is not None:
            url = url._replace(scheme=scheme)
        # check if URL parameters were provided via path
        if url.netloc == '':
            if username is None:
                username = getpass.getuser()
            if hostname is None:
                hostname = socket.gethostname()
            netloc = f'{username}@{hostname}'
            url = url._replace(netloc=netloc)
        # Convert to absolute path
        if url.path.startswith('./') or url.path.startswith('../'):
            path = os.path.abspath(path)
        url = urllib.parse.urlunparse(url)
        self.files.add(url)

    def add_files(self, paths, skip_test=False):
        '''
        Add multiple paths that are associated with an accession

        Parameters
        ----------
        paths : iterable of strings
            The paths the the files
        skip_test : bool
            If true, the method will not test if the file
            exists

        Returns
        -------
        None
        '''
        for path in paths:
            self.add_file(path, skip_test=skip_test)

    def __str__(self):
        return '\n'.join(repr(self).split(','))

    def __repr__(self):  # pragma: no cover
        '''
        String representation of Accession
        '''
        return f'Accession({self.name}, files={self.files}, {self.metadata})'

    @staticmethod
    async def _check_file(url):
        '''
        asyncronously checks a URL
        based in its scheme
        '''
        # Parse the URL and connect
        url = urllib.parse.urlparse(url)
        async with asyncssh.connect(
                url.hostname,
                username=url.username) as conn:
            return await conn.run(
                f'[[ -f {url.path} ]] && echo -n "Y" || echo -n "N"'
            )


    def _check_files(self):
        '''
        Check to see if files attached to an accession are 
        accessible through ssh

        Parameters
        ----------
        None

        Returns
        -------
        Returns True if all files are accessible, otherwise 
        returns a list of files that were unreachable.
        '''
        # Set us up the loop
        tasks = [] 
        loop = asyncio.get_event_loop() 
        # loop through the files and create tasks
        files = list(self.files)
        for url in files:
            tasks.append(self._check_file(url))
        tasks = asyncio.gather(*tasks)
        loop.run_until_complete(tasks)
        unreachable = [i for i,r in enumerate(tasks.result()) if r.stdout!='Y']
        if len(unreachable) == 0:
            return True
        else:
            return [files[x] for x in unreachable]





#!/usr/bin/env python3

import click
import minus80 as m80


@click.group(epilog=f'Made with Love in St Paul -- Version {m80.__version__}')
@click.option('--debug/--no-debug', default=False)
def cli(debug):
    '''
    \b
        __  ____                  ____  ____
       /  |/  (_)___  __  _______( __ )/ __ \\
      / /|_/ / / __ \/ / / / ___/ __  / / / /
     / /  / / / / / / /_/ (__  ) /_/ / /_/ /
    /_/  /_/_/_/ /_/\__,_/____/\____/\____/


    Minus80 is a library for storing biological data. See minus80.linkage.io
    for more details.
    '''

#----------------------------
#    Available Commands
#----------------------------
@click.command(short_help='List the available minus80 datasets',
    help='Reports the available datasets **Frozen** in the minus80 database.'
)
@click.option('--name',  default='*',
    help="The name of the dataset you want to check is available. The default value is the wildcard '*' which will return all available datasets with the specified dtype."
)
@click.option('--dtype', default='',
    help='Each dataset has a datatype associated with it. E.g.: `Cohort`. If no dtype is specified, all available dtypes  will be returned.'
)
def available(name, dtype):
    m80.Tools.available(name, dtype)

cli.add_command(available)

#----------------------------
#    delete Commands
#----------------------------
@click.command(help='Delete a minus80 dataset')
@click.option('--name',help='The name of the dataset to delete')
@click.option('--dtype',help='The dtype of the dataset to delete')
def delete(name, dtype):
    m80.Tools.delete(name,dtype)
cli.add_command(delete)

#----------------------------
#    Cloud Commands
#----------------------------
@click.group()
@click.option('--engine', default='s3', help='Cloud engine.')
@click.option('--raw/--no-raw', default=False, help='Flag to list raw data')
@click.option('--name', default=None, help='Name of m80 dataset')
@click.option('--dtype', default=None, help='Type of m80 dataset')
@click.pass_context
def cloud(ctx, engine, raw, name, dtype):
    '''
    Manage your minus80 datasets in the cloud.
    '''
    ctx.obj = {}
    ctx.obj['engine'] = engine
    ctx.obj['RAW'] = raw
    ctx.obj['NAME'] = name
    ctx.obj['DTYPE'] = dtype

cli.add_command(cloud)

@click.command()
@click.pass_context
def available(ctx):
    '''List available datasets'''
    cloud = m80.CloudData()
    raw = ctx.obj['RAW']
    cloud.list(
        raw=ctx.obj['RAW'],
        name=ctx.obj['NAME'],
        dtype=ctx.obj['DTYPE']
    )


@click.command()
def get():
    'Download a dataset from the cloud'
    pass

@click.command()
def put():
    'Upload a dataset in the cloud'
    pass

cloud.add_command(available)
cloud.add_command(get)
cloud.add_command(put)



import gzip
import bz2
import lzma

class RawFile(object):
    def __init__(self,filename):
        self.filename = filename
        if filename.endswith('.gz'):
            self.handle = gzip.open(filename,'rt')
        elif filename.endswith('bz2'):
            self.handle = bz2.open(filename,'rt')
        elif filename.endswith('xz'):
            self.handle = lzma.open(filenaem,'rt')
        else:
            self.handle = open(filename,'r')
    def __enter__(self):
        return self.handle
    def __exit__(self,dtype,value,traceback):
        self.handle.close()



#!/usr/bin/env python3

from setuptools import setup, find_packages
import os

import io
import re

def read(*names, **kwargs):
    with io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8")
    ) as fp:
        return fp.read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

setup(
    name = 'minus80',
    version = find_version('minus80','__init__.py'),#+'-dev3',
    description = 'A library for freezing, unfreezing and storing biological data.',
    url = 'http://linkage.io',
    author = 'Rob Schaefer',
    license = "Copyright Linkage Analytics 2016. Available under the MIT License",

    classifiers=[
	# How mature is this project? Common values are
	#   3 - Alpha
	#   4 - Beta
	#   5 - Production/Stable
	'Development Status :: 4 - Beta',

	# Indicate who your project is intended for
	'Intended Audience :: Developers',
	'Topic :: Software Development :: Build Tools',

	# Pick your license as you wish (should match "license" above)
	 'License :: OSI Approved :: MIT License',

	# Specify the Python versions you support here. In particular, ensure
	# that you indicate whether you support Python 2, Python 3 or both.
	'Programming Language :: Python :: 3',
	'Programming Language :: Python :: 3.6',
    ],
    keywords='data storage biology freeze', 
    project_urls={
        'Documentation' : 'http://linkage.io/docs/minus80',
        'Source' : 'https://github.com/LinkageIO/Minus80',
        'Tracker' : 'https://github.com/LinkageIO/Minus80/issues'
    },

    packages = find_packages(),
    scripts = [
    ],
    ext_modules = [],
    cmdclass = {
    },

    package_data = {
        '':['*.cyx']    
    },
    install_requires = [		
        'pandas >= 0.22.0',
        'bcolz >= 1.2.1',
        'blaze >= 0.10.1',
        'termcolor >= 1.1.0',
        'pyyaml >= 3.12',
        'click >= 6.7',
        'asyncssh >= 1.12.2',
        'apsw'
    ],
    dependency_links = [
        'git+https://github.com/rogerbinns/apsw'
    ],
    include_package_data=True,
    entry_points='''
        [console_scripts]
        minus80=minus80.cli.minus80:cli
    ''',

    author_email = 'rob@linkage.io',
)



from functools import lru_cache

from minus80 import Accession, Freezable


class Cohort(Freezable):
    '''
        A Cohort is a named set of accessions. Once cohorts are
        created, they are persistant as they are stored in the
        disk by minus80.
    '''

    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self._initialize_tables()

    def __repr__(self):
        return f'Cohort("{self.name}") -- contains {len(self)}'

    @classmethod
    def from_yaml(cls, name, yaml_file):
        '''
        Create a Cohort from a YAML file. Note: this yaml file
        must be created from

        Parameters
        ----------
        name : str
            The name of the Cohort
        yaml_file : pathlike
            The path to the YAML file that contains the
            Accessions

        Returns
        -------
        A Cohort object
        '''
        import yaml
        self = cls(name)
        accessions = yaml.load(open(yaml_file, 'r'))
        self.add_accessions(accessions)
        return self

    @classmethod
    def from_accessions(cls, name, accessions):
        '''
        Create a Cohort from an iterable of Accessions.

        Parameters
        ----------
        name : str
            The name of the Cohort
        accessions : iterable of Accessions
            The accessions that will be frozen in the cohort
            under the given name

        Returns
        -------
        A Cohort object

        '''
        self = cls(name)
        self.add_accessions(accessions)
        return self

    def _initialize_tables(self):
        cur = self._db.cursor()
        cur.execute('''
            CREATE TABLE IF NOT EXISTS accessions (
                AID INTEGER PRIMARY KEY AUTOINCREMENT,
                name NOT NULL UNIQUE
            );
        ''')
        cur.execute('''
            CREATE TABLE IF NOT EXISTS aliases (
                alias TEXT,
                AID INTEGER,
                FOREIGN KEY(AID) REFERENCES accessions(AID)
            );
        ''')
        cur.execute('''
            CREATE TABLE IF NOT EXISTS files (
                AID INTEGER,
                path TEXT ID NOT NULL UNIQUE,
                FOREIGN KEY(AID) REFERENCES accessions(AID)
            );
        ''')
        cur.execute('''
            CREATE TABLE IF NOT EXISTS metadata (
                AID NOT NULL,
                key TEXT NOL NULL,
                val TEXT NOT NULL,
                FOREIGN KEY(AID) REFERENCES accessions(AID)
                UNIQUE(AID, key, val)
            );
        ''')

    @lru_cache(maxsize=2048)
    def _get_AID(self, name):
        '''
            Return a Sample ID (AID)
        '''
        if isinstance(name, Accession):
            name = name.name
        cur = self._db.cursor()
        try:
            return cur.execute(
                'SELECT AID FROM accessions WHERE name = ?', (name, )
            ).fetchone()[0]
        except TypeError:
            pass
        try:
            return cur.execute(
                'SELECT AID FROM aliases WHERE alias = ?', (name, )
            ).fetchone()[0]
        except TypeError:
            raise NameError(f'{name} not in Cohort')

    def random_accession(self):
        '''
            Returns a random accession from the Cohort

            Parameters
            ----------
            None

            Returns
            -------
            Accession
                An Accession object
        '''
        name = self._db.cursor().execute('''
            SELECT name from accessions ORDER BY RANDOM() LIMIT 1;
        ''').fetchone()[0]
        return self[name]

    def random_accessions(self, n=1, replace=False):
        '''
            Returns a list of random accessions from the Cohort, either
            with or without replacement.

            Parameters
            ----------
            n : int
                The number of random accessions to retrieve
            replace: bool
                If false, randomimzation does not include replacement
        '''
        if replace is False:
            if n > len(self):
                raise ValueError(
                        f'Only {len(self)} accessions in cohort. Cannot'
                        ' get {n} samples. See replace parameter in help.'
                )
            return (
                self[name] for (name, ) in self._db.cursor().execute('''
                    SELECT name from accessions ORDER BY RANDOM() LIMIT ?;
                ''', (n, ))
            )
        else:
            return (self.get_random_accession() for _ in range(n))

    def add_accessions(self, accessions):
        '''
            Add multiple Accessions at once
        '''
        with self.bulk_transaction() as cur:
            # When a name is added, it is automatically assigned an ID
            cur.executemany('''
                INSERT OR IGNORE INTO accessions (name) VALUES (?)
            ''', [(x.name, ) for x in accessions])
            # Fetch that ID
            AID_map = self.AID_mapping
            # Populate the metadata and files tables
            cur.executemany('''
                INSERT OR REPLACE INTO metadata (AID, key, val)
                VALUES (?, ?, ?)
            ''', (
                    (AID_map[accession.name], k, v)
                    for accession in accessions
                    for k, v in accession.metadata.items()
                )
            )
            cur.executemany('''
                INSERT OR REPLACE INTO files (AID, path) VALUES (?, ?)
            ''', (
                    (AID_map[accession.name], file)
                    for accession in accessions
                    for file in accession.files
                )
            )
        return [self[x] for x in accessions]

    def add_accession(self, accession):
        '''
            Add a sample to the Database
        '''
        with self.bulk_transaction() as cur:
            # When a name is added, it is automatically assigned an ID
            cur.execute('''
                INSERT OR IGNORE INTO accessions (name) VALUES (?)
            ''', (accession.name, ))
            # Fetch that ID
            AID = self._get_AID(accession)
            # Populate the metadata and files tables
            cur.executemany('''
                INSERT OR REPLACE INTO metadata (AID, key, val)
                VALUES (?, ?, ?)
            ''', ((AID, k, v) for k, v in accession.metadata.items())
            )
            cur.executemany('''
                INSERT OR REPLACE INTO files (AID, path) VALUES (?, ?)
            ''', ((AID, file) for file in accession.files)
            )
        return self[accession]

    @property
    def AID_mapping(self):
        return {
            x.name: x['AID']
            for x in self
        }

    def __delitem__(self, name):
        '''
            Remove a sample by name (or by composition)
        '''
        # First try
        AID = self._get_AID(name)

        self._db.cursor().execute('''
            DELETE FROM accessions WHERE AID = ?;
            DELETE FROM metadata WHERE AID = ?;
            DELETE FROM files WHERE AID = ?;
        ''', (AID, AID, AID))

    def __getitem__(self, name):
        '''
            Get an accession from the database the pythonic way.

            Paremeters
            ----------
            name : object
                Can be a string, i.e. the name or alias of an Accession,
                it can be an Actual Accession OR the AID which
                is an internal ID for accession
        '''
        AID = self._get_AID(name)
        cur = self._db.cursor()
        metadata = {
            k: v for k, v in cur.execute('''
                SELECT key, val FROM metadata WHERE AID = ?;
                ''', (AID, )
            ).fetchall()
        }
        metadata['AID'] = AID
        files = [x[0] for x in cur.execute('''
                SELECT path FROM files WHERE AID = ?;
            ''', (AID, )
            ).fetchall()
        ]
        return Accession(name, files=files, **metadata)

    def __len__(self):
        return self._db.cursor().execute('''
            SELECT COUNT(*) FROM accessions;
        ''').fetchone()[0]

    def __iter__(self):
        for name in (x[0] for x in self._db.cursor().execute('''
                SELECT name FROM accessions
                ''').fetchall()):
            yield self[name]

    def __contains__(self, item):
        if isinstance(item, Accession):
            name = item.name
        else:
            name = item
        try:
            self._get_AID(name)
        except NameError:
            return False
        else:
            return True



from .Config import cf
import os
import shutil

from glob import glob
from collections import defaultdict
from pprint import pprint

__all__ = ['available', 'delete']

def get_files(name, dtype=None, fullpath=False):
    '''
        List the files in the minus80 directory
        associated with a dtype and a name.

        Parameters
        ----------
        name: str, required
            The name of the dataset. Note: accepts glob arguments.
        dtype: str, default=None
            The data type of the dataset. E.g.: Cohort.
            If None, a wildward will be used to retrieve all
            dtypes with the name will be returned.
        fullpath: bool, default=False
            If true, full paths to files will be returned
            if false, only filenames will be returned.


        .. note:: This will only return top level files which sometimes
                  will be directories.
    '''
    bdir = os.path.expanduser(cf.options.basedir)
    if dtype is not None:
        name = f'{name}.*{dtype}.*'
    else:
        name = f'{name}.*'
    data_dir = os.path.join(bdir, 'databases', name)
    files = sorted(glob(data_dir))
    #if dtype is not None:
    #    files = [x for x in files if x.endswith(f'{dtype}.db')]
    #if name is not None:
    #    files = [x for x in files if x.startswith(f'{name}.')]
    if fullpath:
        files = files
    else:
        files = [os.path.basename(x) for x in files]
    return files

def available(name='*',  dtype=''):
    ''' 
        Reports the available datasets **Frozen** in the minus80
        database.

        Parameters
        ----------
        dtype : str
            Each dataset has a datatype associated with it. E.g.:
            `Cohort`. If no dtype is specified, all available dtypes
            will be returned.
        name : str, default:'*'
            The name of the dataset you want to check is available.
            The default value is the wildcard '*' which will return
            all available datasets with the specified dtype.

        Returns
        -------
        bool, None
            If both dtype and name are specified, a bool is returned
            indiciating if the dataset is available. Otherise a formatted
            table is printed and None is returned.
    '''
    files = get_files(name, dtype)

    bdir = os.path.expanduser(cf.options.basedir)
    print(f'Using basedir: {bdir}')
    # Get the names of the individual datasets
    datasets = defaultdict(list)
    for f in files:
        if f.endswith('.db'):
            x, *_, y = f.replace('.db', '').split('.')
            datasets[y].append(x)
    # If both are specified, return a boolean
    if dtype != '' and name != '*':
        if name in datasets[dtype]:
            return True
        else:
            return False
    # Else, print a formatted table
    if len(files) == 0:
        print("--- Nothing here yet ---")
        return None
    for dtype, names in datasets.items():
        print(f"--- {dtype}: -----------------")
        for i, name in enumerate(names, 1):
            print(f'\t{i}. {name}')

def delete(name, dtype=None, force=False):
    ''' 
        Deletes files associated with Minus80 datasets.

        Parameters
        ----------
        name : str
            The name of the dataset you want to delete
        dtype : str
            Each dataset has a datatype associated with it. E.g.:
            `Cohort`. If no dtype is specified, all available dtypes
            will be returned.
        force : bool, default: False
            If False, the function will list off the files it wants to delete.
            If True, it will do what you tell it to do and just delete things
            (not recommended).

        Returns
        -------
        int
            Returns the number of files deleted

        .. warning:: This is damaging. Deleted datasets cannot be (easily) recovered.
    '''
    # Get a filecard for all the minus80 filenames that match the
    # type and the name
    files = get_files(name, dtype=dtype)
    if force != True:
        print(f'Are you sure you want to delete {len(files)} files?:\n')
        pprint(f'{files}')
        if input('[y/n]: ').upper() != 'Y':
            print('Nothing deleted.')
            return 0
    # delete them
    num_deleted = 0
    for filename in files:
        bdir = os.path.expanduser(cf.options.basedir)
        data_dir = os.path.join(bdir, 'databases')
        filename = os.path.join(data_dir, filename)
        if os.path.isfile(filename):
            os.remove(filename)
        elif os.path.isdir(filename):
            shutil.rmtree(filename)
        num_deleted += 1
    return num_deleted


def directory_search(directory, suffix='.fastq'):
    '''
        Search a directory and create Accessions from the files found there.
    '''
    from minus80 import Accession

    accessions = dict()
    for root, _, files in os.walk(directory, followlinks=True):
        for f in files:
            if f.endswith(suffix):
                fid, *_ = f.split('_')
                if fid not in accessions:
                    accessions[fid] = Accession(fid, files=[])
                x = accessions[fid]
                x.add_file(os.path.join(root, f))


    return accessions




